==> java-gnome-4.0.3.tar.gz: java-gnome-4.0.3/src/jni/bindings_java_memory.c <==
/*
 * bindings_java_memory.c
 *
 * Copyright (c) 2007 Operational Dynamics Consulting Pty Ltd and Others
 * 
 * The code in this file, and the library it is a part of, are made available
 * to you by the authors under the terms of the "GNU General Public Licence,
 * version 2" plus the "Classpath Exception" (you may link to this code as a
 * library into other programs provided you don't make a derivation of it).
 * See the LICENCE file for the terms governing usage and redistribution.
 */

/*
 * We would be remiss if we did not record that the GObject ToggleRef
 * capability was worked out by Owen Taylor especially for language bindings
 * and that the original implementation in the old java-gnome 2.x was done by
 * Nick Rahn, Ismael Juma, and Jeffrey Morgan, ultimately coming to rest in
 * glib-java/src/jni/glib_java.c. Although our overall Proxy strategy, memory
 * management approach, and use of ToggleRef is somewhat different, the idea
 * of switching to a weak Java reference (rather than dropping it entirely)
 * is inspired from that code.
 */

#include <jni.h>
#include <glib.h>
#include <glib-object.h>
#include "bindings_java.h"

#define REFERENCE "RefToJavaProxy"

/**
 * This function is called whenever the GObject reference count drops to one,
 * and that one reference is the ToggleRef. When that happens, we drop the
 * strong Java reference which means that if that was the last such strong
 * reference, the Java object can be finalized (with the result that the
 * ToggleRef will in turn be dropped, allowing the GObject to destroy).
 * 
 * We switch to a weak Java reference instead of just removing it so that we
 * can find our way back to the Java object should the ToggleRef be activated
 * with a "you're not the last ref anymore".
 */
/*
 * Signature the prototype of a (*GToggleNotify) callback, meeting the
 * requirements of the second argument to g_object_add_toggle_ref()
 */
static void
bindings_java_toggle
(
	gpointer data,
        GObject *object,
        gboolean is_last_ref
)
{
	JNIEnv* env;
	jobject ref;
	jobject strong;
	jobject weak;
	
	ref = g_object_get_data(object, REFERENCE);
	env = bindings_java_getEnv();
	
	if (is_last_ref) {
		/*
		 * Create weak Java reference, replace association stored in
		 * GObject, and remove strong Java reference
		 */
		if (DEBUG_MEMORY_MANAGEMENT) {
			g_print("mem: toggle Java ref to WEAK\t%ld\n", (long) object);
		}
		weak = (*env)->NewWeakGlobalRef(env, ref);
		g_object_set_data(object, REFERENCE, weak);
 		(*env)->DeleteGlobalRef(env, ref);
	} else {
		/*
		 * Create strong Java reference, overwrite association in
		 * GObject, and remove weak Java reference now that we've
		 * replaced it with a strong one. 
		 */
		if (DEBUG_MEMORY_MANAGEMENT) {
			g_print("mem: toggle Java ref to STRONG\t%ld\n", (long) object);
		}
		strong = (*env)->NewGlobalRef(env, ref);
		g_object_set_data(object, REFERENCE, strong);
		(*env)->DeleteWeakGlobalRef(env, ref);
	}	
}

/**
 * Drop our initial default GObject reference. This is called via a GIdle
 * call to avoid the ping pong effect which results if the ToggleRef is the
 * only ref count to a GtkWidget while setters are called prior to the
 * GtkWidget being added to a GtkContainer. By delaying the ref count
 * reduction, the initial strong Java reference is left alone.  
 */
/*
 * Signature the prototype of a (*GSourceFunc) callback, meeting the
 * requirements of the first argument to g_idle_add()
 */
static gboolean
bindings_java_memory_deref
(
        gpointer data
)
{
        GObject* object;
        
        object = (GObject*) data;
        g_object_unref(object);
        return FALSE;
}       


/**
 * Called from
 *   Java_org_gnome_glib_GObject_g_1object_1add_1toggle_1ref()
 * called from
 *   org.gnome.glib.GObject.g_object_add_toggle_ref(long reference, Object target)
 * called from
 *   org.gnome.glib.GObject.addToggleRef(Object reference)
 * called from
 *   org.gnome.glib.Object.<init>(long pointer)
 *
 * Here we go through the various motions necessary to appropriately
 * increment the ref count on the GObject for which we are creating a Proxy
 * to.
 */
void
bindings_java_memory_ref
(
	JNIEnv* env,
	GObject* object,
	jobject target 
)
{
	jobject strong;

	/*
	 * Take out a strong Java reference "from the GObject to the Proxy"
	 * and stash it in the GObject. 
	 */
 
 	if (DEBUG_MEMORY_MANAGEMENT) {
 		g_print("mem: add STRONG Java ref\t%ld\n", (long) object);
 	}
	strong = (*env)->NewGlobalRef(env, target);
	g_object_set_data(object, REFERENCE, strong);

	/*
	 * Add a ToggleRef to the GObject "from the Proxy to the GObject",
	 * incrementing its ref count
	 */

	g_object_add_toggle_ref(object, bindings_java_toggle, NULL);

	/*
	 * All GObjects are created with an initial ref, owned by the caller.
	 * The ToggleRef we just created makes a second ref. We need to drop
	 * the original.
	 *
	 * In GTK, most objects are subclasses of GtkObject which are 
	 * derived from GInitiallyUnowned, marking them to be created with
	 * a "floating" ref for better API convenience when working in C. That
	 * reference must be "sunk" (us taking ownership of the Ref), then
	 * likewise dropped.
	 *
	 * In some very rare situations, however, GTK will have sunk the Ref
	 * already (the primary example is GtkWindow for which GTK sinks the
	 * reference and uses it to connect to the X server) in which case we
	 * must NOT drop the Ref we don't own. 
	 *
	 * To prevent memory leaks, we must take care about all these
	 * situations.
	 */
	if (G_IS_INITIALLY_UNOWNED(object)) {
		if (!g_object_is_floating(object)) {
			return;
		}
		g_object_ref_sink(object);
	}
	g_idle_add(bindings_java_memory_deref, object);
}
	
/**
 * Called from
 *   Java_org_gnome_glib_GObject_g_1object_1remove_1toggle_1ref()
 * called from
 *   org.gnome.glib.GObject.g_object_remove_toggle_ref(long reference)
 * called from
 *   org.gnome.glib.GObject.removeToggleRef(Object reference)
 * called when
 *   org.gnome.glib.Object.release()
 * is invoked by a finalizer.
 * 
 * Remove the ToggleRef. Since this is called by the Java finalizer, and
 * since the Java object should only become only-weakly-reachable when there
 * are no more GObject refs, this should drop the last ref count to the
 * GObject, resulting it in it being eligible for destruction.  
 */
void
bindings_java_memory_unref
(
	GObject* object
)
{
	if (DEBUG_MEMORY_MANAGEMENT) {
		g_print("mem: remove toggle ref for\t%ld\n", (long) object);
	}
	g_object_remove_toggle_ref(object, bindings_java_toggle, NULL);
}

==> java-gnome-4.0.8.tar.gz: java-gnome-4.0.8/src/jni/bindings_java_memory.c <==
/*
 * bindings_java_memory.c
 *
 * Copyright (c) 2007 Operational Dynamics Consulting Pty Ltd and Others
 * 
 * The code in this file, and the library it is a part of, are made available
 * to you by the authors under the terms of the "GNU General Public Licence,
 * version 2" plus the "Classpath Exception" (you may link to this code as a
 * library into other programs provided you don't make a derivation of it).
 * See the LICENCE file for the terms governing usage and redistribution.
 */

/*
 * We would be remiss if we did not record that the GObject ToggleRef
 * capability was worked out by Owen Taylor especially for language bindings
 * and that the original implementation in the old java-gnome 2.x was done by
 * Nick Rahn, Ismael Juma, and Jeffrey Morgan, ultimately coming to rest in
 * glib-java/src/jni/glib_java.c. Although our overall Proxy strategy, memory
 * management approach, and use of ToggleRef is somewhat different, the idea
 * of switching to a weak Java reference (rather than dropping it entirely)
 * is inspired from that code.
 */

#include <jni.h>
#include <glib.h>
#include <glib-object.h>
#include "bindings_java.h"

#define REFERENCE "RefToJavaProxy"

/**
 * This function is called whenever the GObject reference count drops to one,
 * and that one reference is the ToggleRef. When that happens, we drop the
 * strong Java reference which means that if that was the last such strong
 * reference, the Java object can be finalized (with the result that the
 * ToggleRef will in turn be dropped, allowing the GObject to destroy).
 * 
 * We switch to a weak Java reference instead of just removing it so that we
 * can find our way back to the Java object should the ToggleRef be activated
 * with a "you're not the last ref anymore".
 */
/*
 * Signature the prototype of a (*GToggleNotify) callback, meeting the
 * requirements of the second argument to g_object_add_toggle_ref()
 */
static void
bindings_java_toggle
(
	gpointer data,
        GObject *object,
        gboolean is_last_ref
)
{
	JNIEnv* env;
	jobject ref;
	jobject strong;
	jobject weak;

	ref = g_object_get_data(object, REFERENCE);
	env = bindings_java_getEnv();

	if (is_last_ref) {
		/*
		 * Create weak Java reference, replace association stored in
		 * GObject, and remove strong Java reference
		 */
		if (DEBUG_MEMORY_MANAGEMENT) {
			g_print("mem: toggle Java ref to WEAK\t%s\n", bindings_java_memory_pointerToString(object));
		}
		weak = (*env)->NewWeakGlobalRef(env, ref);
		g_object_set_data(object, REFERENCE, weak);
 		(*env)->DeleteGlobalRef(env, ref);
	} else {
		/*
		 * Create strong Java reference, overwrite association in
		 * GObject, and remove weak Java reference now that we've
		 * replaced it with a strong one. 
		 */
		if (DEBUG_MEMORY_MANAGEMENT) {
			g_print("mem: toggle Java ref to STRONG\t%s\n", bindings_java_memory_pointerToString(object));
		}

		strong = (*env)->NewGlobalRef(env, ref);
		g_object_set_data(object, REFERENCE, strong);
		(*env)->DeleteWeakGlobalRef(env, ref);
	}	
}

/**
 * Drop our initial default GObject reference. This is called via a GIdle
 * call to avoid the ping pong effect which results if the ToggleRef is the
 * only ref count to a GtkWidget while setters are called prior to the
 * GtkWidget being added to a GtkContainer. By delaying the ref count
 * reduction, the initial strong Java reference is left alone.  
 */
/*
 * Signature the prototype of a (*GSourceFunc) callback, meeting the
 * requirements of the first argument to g_idle_add()
 */
static gboolean
bindings_java_memory_deref
(
        gpointer data
)
{
        GObject* object;

        object = (GObject*) data;
        
                
	if (DEBUG_MEMORY_MANAGEMENT) {
		g_print("mem: drop GObject ref\t\t%s\n", bindings_java_memory_pointerToString(object));
	}
        g_object_unref(object);
        return FALSE;
}       

/**
 * Called from
 *   Java_org_gnome_glib_GObject_g_1object_1add_1toggle_1ref()
 * called from
 *   org.gnome.glib.GObject.g_object_add_toggle_ref(long reference, Object target)
 * called from
 *   org.gnome.glib.GObject.addToggleRef(Object reference)
 * called from
 *   org.gnome.glib.Object.<init>(long pointer)
 *
 * Here we go through the various motions necessary to appropriately
 * increment the ref count on the GObject for which we are creating a Proxy
 * to.
 */
void
bindings_java_memory_ref
(
	JNIEnv* env,
	GObject* object,
	jobject target 
)
{
	jobject strong;

	/*
	 * Take out a strong Java reference "from the GObject to the Proxy"
	 * and stash it in the GObject. 
	 */
 
 	if (DEBUG_MEMORY_MANAGEMENT) {
 		g_print("mem: add STRONG Java ref\t%s\n", bindings_java_memory_pointerToString(object));
 	}
	strong = (*env)->NewGlobalRef(env, target);
	g_object_set_data(object, REFERENCE, strong);

	/*
	 * Add a ToggleRef to the GObject "from the Proxy to the GObject",
	 * incrementing its ref count
	 */

	g_object_add_toggle_ref(object, bindings_java_toggle, NULL);

	/*
	 * And drop the reference we originally own. Note that our automatically
	 * generated code ensure we always own a ref, adding it if we don't have
	 * such reference originally.
	 */
	g_idle_add(bindings_java_memory_deref, object);
}


/**
 * Called from
 *   Java_org_gnome_glib_GObject_g_1object_1remove_1toggle_1ref()
 * called from
 *   org.gnome.glib.GObject.g_object_remove_toggle_ref(long reference)
 * called from
 *   org.gnome.glib.GObject.removeToggleRef(Object reference)
 * called when
 *   org.gnome.glib.Object.release()
 * is invoked by a finalizer.
 * 
 * Remove the ToggleRef. Since this is called by the Java finalizer, and
 * since the Java object should only become only-weakly-reachable when there
 * are no more GObject refs, this should drop the last ref count to the
 * GObject, resulting it in it being eligible for destruction.  
 */
void
bindings_java_memory_unref
(
	GObject* object
)
{
	if (DEBUG_MEMORY_MANAGEMENT) {
		g_print("mem: remove toggle ref for\t%s\n", bindings_java_memory_pointerToString(object));
	}

	g_object_remove_toggle_ref(object, bindings_java_toggle, NULL);
}

void
bindings_java_memory_cleanup
(
    GObject* object,
    gboolean owner
)
{
    if (g_object_get_data(object, REFERENCE) == NULL) {
        
        /* 
         * We don't have a proxy for the given object. If we own the object,
         * we don't need to do anything, as Object constructor will take care
         * of adding the toggle ref and removing the ref we own. There is an
         * exception, however, as for GInitiallyUnowned we need to sink the
         * reference if we have one.
         * 
         * If we don't own the object, we need to add a new ref, that will be
         * later removed by the Proxy constructor.
         */
        if (owner) {
            if (G_IS_INITIALLY_UNOWNED(object) && g_object_is_floating(object)) {
                if (DEBUG_MEMORY_MANAGEMENT) {
                    g_print("mem: sink GObject ref\t\t%s\n", bindings_java_memory_pointerToString(object));
                }
                g_object_ref_sink(object);
            }
        } else {
            /*
             * We do not own the object, so we need to add an extra ref, as
             * Object constructor assumes we actually own the object.
             */
            if (DEBUG_MEMORY_MANAGEMENT) {
                g_print("mem: added extra ref for\t%s\n", bindings_java_memory_pointerToString(object));
            }
            g_object_ref(object);
        }
    } else {
        /*
         * We already have a proxy for the object. So we already have our 
         * toggle ref. Thus, if a methods adds an extra ref, we need to drop
         * it. 
         * TODO does such behavior exist in Gtk+?
         */
        if (owner) {
            if (DEBUG_MEMORY_MANAGEMENT) {
                g_print("mem: remove ref for\t%s\n", bindings_java_memory_pointerToString(object));
            }
            g_object_unref(object);
        }
    }
}

#if GLIB_SIZEOF_VOID_P == 8
#define WIDTH "16"
#define SIZE GLIB_SIZEOF_VOID_P * 2 + 3
#else
#define WIDTH "8"
#define SIZE GLIB_SIZEOF_VOID_P * 2 + 3
#endif

/*
 * A utility function to format a pointer address as an appropriate width
 * string. This is necessary because of the inadequecies of printf's %p.
 * The return value is statically allocated and must not be free'd. 
 * 
 * This is called when debugging, both from C and from Java via
 * Plumbing.toHexString()
 */
const gchar*
bindings_java_memory_pointerToString
(
	gpointer pointer
)
{
	static gchar result[SIZE];
	
	g_snprintf(result, SIZE, "0x%." WIDTH "lX", (unsigned long) pointer);
	
	return result;
}
